from django.shortcuts import render
from django.http import HttpResponseRedirect, HttpResponse
from django.urls import reverse
from .models import PCTLProperty
from django.template import loader
from django.shortcuts import get_object_or_404,render
from django.http import Http404
from django.views import generic
from .forms import PCTLPropertyForm
from django.conf import settings
import os
from verification import Verification
from django.utils.encoding import smart_str
from wsgiref.util import FileWrapper 
import mimetypes

class IndexView(generic.ListView):
    context_object_name = 'pctl_properties_list'

    def get_queryset(self):
        return PCTLProperty.objects.all()


def update(request, pctl_property_id):
	if request.method == 'POST':
		if("dashboard_update" in request.POST):
			pctl_property = get_object_or_404(PCTLProperty, id=pctl_property_id)
			pctl_property.pctl_text = request.POST.get('pctl_text','')
			pctl_property.elasticity_restrictions = request.POST.get('restrictions','add,remove,no_op')
			pctl_property.save()
			return render(request, 'analysis/verification.html', {
				'pctl_property': pctl_property,
				'elasticity_manager': pctl_property.elasticity_manager_id,
			})
		else:
			instance = get_object_or_404(PCTLProperty, id=pctl_property_id)
			form = PCTLPropertyForm(request.POST or None, instance=instance)
			form.save()
			return HttpResponseRedirect(reverse('analysis:pctl_index'))
	else:
		pctl_property = get_object_or_404(PCTLProperty, id=pctl_property_id)
		form = PCTLPropertyForm(instance=pctl_property) #we need the instance to update
		return render(request, 'analysis/details.html', {
        		'form': form,
			'update': True,
			'pctl_property':pctl_property,
    		})

def create(request):
    if request.method == 'POST':
	form = PCTLPropertyForm(request.POST)
	form.save()
	return HttpResponseRedirect(reverse('analysis:pctl_index'))
    else:
	form = PCTLPropertyForm()
	return render(request, 'analysis/details.html', {
        	'form': form,
    	})

def delete(request, pctl_property_id):
    pctl_property = get_object_or_404(PCTLProperty, pk=pctl_property_id)
    pctl_property.delete()
    return HttpResponseRedirect(reverse('analysis:pctl_index'))

def verify(request, pctl_property_id):
	pctl_property = get_object_or_404(PCTLProperty, pk=pctl_property_id)
	past_log = int(request.GET.get('past_log', "-1"))
	version = "long"
	allowed_actions=pctl_property.elasticity_restrictions.split(",")

	if("restrictions" in request.GET):
		allowed_actions = request.GET.get('restrictions', "").split(',')
	
	if(request.GET.get('version', "long")):
		version = request.GET.get('version', "long")
	properties_file = os.path.join(settings.BASE_DIR, 'decision')+"/exp_"+str(pctl_property.elasticity_manager_id)+"/Coordinator_"+str(pctl_property.elasticity_manager_id)+".properties"
	solver = Verification.Verification(properties_file,past_log)
	out = solver.verify(pctl_property.pctl_text,shortVersion=(version=="short"),allowed_actions=allowed_actions)
	return HttpResponse("<pre>"+out+"</pre>")

def verification_dashboard(request, pctl_property_id):
	pctl_property = get_object_or_404(PCTLProperty, id=pctl_property_id)
	return render(request, 'analysis/verification.html', {
		'pctl_property': pctl_property,
		'elasticity_manager': pctl_property.elasticity_manager_id,
	})

def download_model(request, pctl_property_id):
    pctl_property = get_object_or_404(PCTLProperty, pk=pctl_property_id)
    file_path = os.path.join(settings.BASE_DIR, 'decision')+"/exp_"+str(pctl_property.elasticity_manager_id)+"/prism_output_"+str(pctl_property.elasticity_manager_id)+"/model.prism"
    if os.path.exists(file_path):
	wrapper = FileWrapper( open( file_path, "r" ) )
	content_type = mimetypes.guess_type( file_path )[0]

	response = HttpResponse(wrapper, content_type = content_type)
	response['Content-Length'] = os.path.getsize( file_path ) # not FileField instance
	response['Content-Disposition'] = 'attachment; filename=' + smart_str( os.path.basename( file_path ) )
	return response
    else:
        raise Http404
