import logging, math, time, thread, random, itertools, sys
import Utils, operator, os, datetime
import numpy as np
import pandas
import statsmodels.api as sm
from scipy.cluster.vq import vq, kmeans, kmeans2, whiten
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import centroid, linkage, fcluster, fclusterdata
from scipy.stats import linregress
from collections import deque
from collections import defaultdict;
from PrismHandler import PrismHandler
from SpotHandler import SpotHandler
from random import randrange
import My_Logger
from collections import deque
from sklearn.linear_model import LinearRegression
from pysqlite2 import dbapi2 as sqlite

class MDPDecisionMaker_sim():
    def __init__(self, properties_file):
        self.utils = Utils.Utils(properties_file)
	self.spotHandler = SpotHandler(properties_file)
	self.experiment_log = My_Logger.My_Logger(properties_file);
        self.polManager =  PolicyManager("test",properties_file)
        self.acted = ["done"]
        cluster_size = self.utils.initial_cluster_size
        self.currentState = str(cluster_size)
        self.nextState = str(cluster_size)
	self.vm_types_currentState = self.utils.vm_types_currentState
	self.vm_types_nextState = self.utils.vm_types_currentState
        self.debug = False
        self.prediction = False
        self.waitForIt = 0
        # The policy for getting throughput and latency when computing the reward func.
        # average, centroid
        self.measurementsPolicy = 'centroid'
        self.countdown = 0
	self.n_current = 50
	self.past_lambda = np.array([])

	self.latencies = np.array([])

	#window configuration
	if eval(self.utils.window_inlambda):
		self.latestMeasQueue = deque([])
		self.windowSize = 100
	elif eval(self.utils.ewma_inlambda):
		self.lambdaQueue = deque()
		self.latQueue = deque()
		self.thrQueue = deque()
		self.windowSize = 100


	##Create model
	if not os.path.exists(self.utils.prism_output_folder):
    		os.makedirs(self.utils.prism_output_folder);
	
	self.mdp_handler = PrismHandler(properties_file)
        # A dictionary that will remember rewards and metrics in states previously visited
        self.memory = {}
        
        for i in range(int(self.utils.min_cluster_size), int(self.utils.max_cluster_size)+1):
            self.memory[str(i)] = {}
            self.memory[str(i)]['V'] = None # placeholder for rewards and metrics
            self.memory[str(i)]['r'] = None
            self.memory[str(i)]['arrayMeas'] = None
        
        ## Install logger
        LOG_FILENAME = self.utils.install_dir+'/'+self.utils.logs_folder_name+'/Coordinator.log'
        self.my_logger = logging.getLogger('MDPDecisionMaker_sim')
        self.my_logger.setLevel(self.utils.logging_level)
        
        handler = logging.handlers.RotatingFileHandler(
                      LOG_FILENAME, maxBytes=2*1024*1024, backupCount=5)
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
        handler.setFormatter(formatter)
        self.my_logger.addHandler(handler)
        
        # Load any previous statics.
        self.measurementsFile = self.utils.install_dir+'/'+self.utils.logs_folder_name+'/measurements.txt'
        self.sumMetrics = {}
        meas = open(self.measurementsFile, 'a')
        if os.stat(self.measurementsFile).st_size == 0:
            # The file is empty, set the headers for each column.
            meas.write('State\t\tLambda\t\tThroughput\t\tLatency\t\tCPU\t\tTime\n')
        else :
            # Read the measurements saved in the file.
            meas.close()
            meas = open(self.measurementsFile, 'r')
            
            meas.next() # Skip the first line with the headers of the columns
            for line in meas:
                # Skip comments (used in training sets)
                if not line.startswith('###'):
                    m = line.split('\t\t')
                    self.addMeasurement(m)
        
        meas.close()

	con = sqlite.connect(self.utils.db_file)
	cur = con.cursor()
	cur.execute('drop table if exists instances')
	con.commit()
	cur.execute('create table instances(id int, launch_time text, vm_type text)')
	con.commit()
	for vm_type in self.utils.vm_types_currentState:
		for i in range(0,int(self.utils.vm_types_currentState[vm_type])):
			cur.execute(""" insert into instances(id, launch_time, vm_type) values  (?,?,?)""", (i, str(time.time()),vm_type))
			con.commit()
	cur.close()
	con.close()

	self.cost_before_remove = 0;
	self.removed_vms_cnt = 0;
               
    # param metrics: array The metrics to store. An array containing [state, lamdba, throughput, latency, cpu, time]
    # param writeFile: boolean If set write the measurement in the txt file
    def addMeasurement(self, metrics, writeFile=False):
	#store all the latency meas to compute mean and stdev for z-score normalization
	self.latencies = np.append(self.latencies,[float(metrics[3])])

        if self.measurementsPolicy.startswith('average'):
            if not self.sumMetrics.has_key(metrics[0]):
                # Save the metric with the state as key metrics = [state, inlambda, throughput, latency]
                self.sumMetrics[metrics[0]] = {'inlambda': 0.0, 'throughput': 0.0, 'latency': 0.0, 'divide_by': 0}
            
            self.sumMetrics[metrics[0]] = {'inlambda': self.sumMetrics[metrics[0]]['inlambda'] + float(metrics[1]),
                                           'throughput': self.sumMetrics[metrics[0]]['throughput'] + float(metrics[2]), 
                                     'latency': self.sumMetrics[metrics[0]]['latency'] + float(metrics[3]),
                                     'divide_by': self.sumMetrics[metrics[0]]['divide_by'] + 1}

        # metrics-> 0: state, 1: lambda, 2: thoughtput, 3:latency, 4:cpu, 5:time
        if not self.memory.has_key(metrics[0]):
            self.memory[str(metrics[0])] = {}
            self.memory[str(metrics[0])]['V'] = None # placeholder for rewards and metrics
            self.memory[str(metrics[0])]['r'] = None
            self.memory[str(metrics[0])]['arrayMeas'] = np.array([float(metrics[1]), float(metrics[2]), 
                                                                      float(metrics[3]), float(metrics[4])], ndmin=2)
        elif self.memory[metrics[0]]['arrayMeas'] == None:
            self.memory[metrics[0]]['arrayMeas'] = np.array([float(metrics[1]), float(metrics[2]), 
                                                                     float(metrics[3]), float(metrics[4])], ndmin=2)
        else:
            self.memory[metrics[0]]['arrayMeas'] = np.append(self.memory[metrics[0]]['arrayMeas'], 
                                                             [[float(metrics[1]), float(metrics[2]), 
                                                               float(metrics[3]), float(metrics[4])]], axis=0)
         
        if writeFile:
	    #update window information    
	    if eval(self.utils.window_inlambda):
		if len(self.latestMeasQueue) > self.windowSize:
			self.latestMeasQueue.popleft()
		self.latestMeasQueue.append([float(metrics[1]), float(metrics[2]), float(metrics[3]), float(metrics[4])])
	    elif eval(self.utils.ewma_inlambda):
		if len(self.lambdaQueue) > self.windowSize:
			self.lambdaQueue.popleft()
			self.thrQueue.popleft()
			self.latQueue.popleft()
		self.lambdaQueue.append(float(metrics[1]))
		self.thrQueue.append(float(metrics[2]))
		self.latQueue.append(float(metrics[3]))

	    #keep track of all the past lambda for prediction
	    self.past_lambda = np.append(self.past_lambda,[float(metrics[1])]) 

	    vm_types_currentState_logs = ""
	    for vm_type in self.vm_types_currentState:
		vm_types_currentState_logs += '\t\t'+str(self.vm_types_currentState[vm_type])+'@'+vm_type

            ms = open(self.measurementsFile, 'a')
            # metrics[5] contains the time tick -- when running a simulation, it represents the current minute, 
            # on actual experiments, it is the current time. Used for debugging and plotting
            ms.write(str(metrics[0])+'\t\t'+str(metrics[1])+'\t\t'+str(metrics[2])+'\t\t'+
                     str(metrics[3])+'\t\t'+str(metrics[4])+'\t\t'+str(metrics[5])+vm_types_currentState_logs+'\n')
            ms.close()
	    
	    meas_dict =  {'num_nodes': float(metrics[0]),'inlambda': metrics[1],'throughput': metrics[2],'latency': metrics[3],'cpu': metrics[4]}       
	    meas_dict['max_num_nodes'] = int(self.utils.max_cluster_size)

	    total_cost,norm_total_cost = self.get_the_cost()
    	    meas_dict['total_cost'] = total_cost
	    meas_dict['norm_total_cost'] = norm_total_cost
	    meas_dict['norm_latency'] = self.get_norm_latency(meas_dict['latency'],np.mean(self.latencies),np.std(self.latencies))
	    if(float(float(meas_dict['latency'])) > float(self.utils.latency_threshold)):
		gain = float(self.utils.gain_punishment);
	    else:
	    	gain = float(eval(self.utils.gain,meas_dict))
	    transition_cost = float(eval(self.utils.trans_cost,meas_dict))   

	    not_attack_probability = 1;
	    for vm_type in self.vm_types_currentState:
		not_attack_probability = not_attack_probability * pow(1-self.utils.vm_types_attack_probs[vm_type],self.vm_types_currentState[vm_type])
	    attack_probability = 1 - not_attack_probability
	    attack_incident = self.generate(attack_probability)           
	    self.experiment_log.log_measurement(str(metrics[0])+','+str(metrics[1])+','+str(metrics[2])+','+str(metrics[3])+','+str(metrics[4])+','+str(gain-transition_cost)+','+str(attack_incident))


	
    # param state: string Get the average metrics (throughput, latency) for this state.
    # return a dictionary with the averages
    def getAverages(self, state):
        averages = {}
        if self.sumMetrics.has_key(state):
            averages['throughput'] = float(self.sumMetrics[state]['throughput']/self.sumMetrics[state]['divide_by'])
            averages['latency'] = float(self.sumMetrics[state]['latency']/self.sumMetrics[state]['divide_by'])
            
            self.my_logger.debug("GETAVERAGES Average metrics for state: "+ state +" num of measurements: "+ str(self.sumMetrics[state]['divide_by']) +
                                 " av. throughput: "+ str(averages['throughput']) +" av. latency: " +
                                 str(averages['latency']))
        return averages
    
    def getCentroids(self):
        centroids = {}
        for i in range(int(self.utils.min_cluster_size), int(self.utils.max_cluster_size)+1):
            if self.memory[str(i)]['arrayMeas'] != None:
                self.my_logger.debug("GETCENTROIDS state "+ str(i) +" measurements : "+ str(self.memory[str(i)]['arrayMeas']))
                if len(self.memory[str(i)]['arrayMeas']) > 1:
                    Y = self.memory[str(i)]['arrayMeas']
                    T= fclusterdata(self.memory[str(i)]['arrayMeas'], t=15.0, criterion='distance', metric='euclidean', method='single')
                    self.my_logger.debug("GETCENTROIDS state "+ str(i) +" clusters: "+ str(T))
                    Z = centroid(Y)
                    self.my_logger.debug("GETCENTROIDS state "+ str(i) +" centroid func: "+ str(Z))
                else:
                    centroids[str(i)] = {}
                    centroids[str(i)]['throughput'] = self.memory[str(i)]['arrayMeas'][0][0]
                    centroids[str(i)]['latency'] = self.memory[str(i)]['arrayMeas'][0][1]
        
        self.my_logger.debug("GETCENTROIDS centroids: "+ str(centroids))
        return centroids
                    
    def doKmeans(self, state, from_inlambda, to_inlambda):
        # Run kmeans for the measurements of this state and return the centroid point (throughput, latency)
        ctd = {}
        label = []
        if self.memory[state]['arrayMeas'] != None :
            self.my_logger.debug("DOKMEANS length of self.memory[state]['arrayMeas']: " + str(len(self.memory[state]['arrayMeas'])))
            sliced_data = None
            for j in self.memory[state]['arrayMeas']:
                #self.my_logger.debug("DOKMEANS self.memory[state]['arrayMeas'][j]: "+ str(j))
                # If this measurement belongs in the slice we're insterested in
                if j[0] >= from_inlambda and j[0] <= to_inlambda:
                    #self.my_logger.debug("DOKMEANS adding measurement : "+ str(j))
                    # add it
                    if sliced_data == None:
                        sliced_data = np.array(j, ndmin=2)
                    else:
                        sliced_data = np.append(sliced_data, [j], axis=0)
            
            k = 4
            # 1. No known lamdba values close to current lambda measurement
            if sliced_data == None:
                self.my_logger.debug("No known lamdba values close to current lambda measurement. Returning zeros!")
		
            else:
                self.my_logger.debug("DOKMEANS length of sliced_data to be fed to kmeans: "+ str(len(sliced_data)))
                centroids, label = kmeans2(sliced_data, k, minit='points')
                
            # initialize dictionary
            num_of_meas = {}
            #num_of_meas = {'0': 0, '1': 0, '2': 0, '3': 0, '4': 0}
            for j in range(0, k):
                num_of_meas[str(j)] = 0
            if len(label) > 0:
                for i in label:
                    num_of_meas[str(i)] += 1 
            
                max_meas_cluster = max(num_of_meas.iteritems(), key=operator.itemgetter(1))[0]
                ctd['inlambda'] = centroids[int(max_meas_cluster)][0]
                ctd['throughput'] = centroids[int(max_meas_cluster)][1]
                ctd['latency'] = centroids[int(max_meas_cluster)][2]
                ctd['cpu'] = centroids[int(max_meas_cluster)][3]
            else:
                self.my_logger.debug("DOKMEANS one of the clusters was empty and so label is None :|. Returning zeros")
                ctd['inlambda'] = 0.0
                ctd['throughput'] = 0.0
                ctd['latency'] = 0.0
                ctd['cpu'] = 0.0
                #return None
        else:
            self.my_logger.debug("DOKMEANS self.memory[state]['arrayMeas'] is None :|")

        return ctd

    def getNewCentroids(self,sliced_data,k,centroids,label):
	inlambda_sum = [0] * k
	inlambda_cnt = [0] * k
	thr_sum = [0] * k
	thr_cnt = [0] * k
	cpu_sum = [0] * k
	cpu_cnt = [0] * k

	i = 0;
	for l in label:
		inlambda_sum[l] = inlambda_sum[l] + sliced_data[i][0]
		inlambda_cnt[l] = inlambda_cnt[l] + 1
		thr_sum[l] = thr_sum[l] + sliced_data[i][1]
		thr_cnt[l] = thr_cnt[l] + 1
		cpu_sum[l] = cpu_sum[l] + sliced_data[i][3]
		cpu_cnt[l] = cpu_cnt[l] + 1
		i = i+1

	for i in range(0,k):
		if inlambda_cnt[i] != 0 and len(centroids[:][i]) != 0:
			if i == 0:
				a = np.array([inlambda_sum[i]/inlambda_cnt[i],thr_sum[i]/thr_cnt[i],centroids[:][i][0],cpu_sum[i]/cpu_cnt[i]], ndmin=2)
			else:
				a = np.append(a, [[inlambda_sum[i]/inlambda_cnt[i],thr_sum[i]/thr_cnt[i],centroids[:][i][0],cpu_sum[i]/cpu_cnt[i]]], axis=0)
		elif inlambda_cnt[i] == 0:
			if i == 0:
				a = np.array([0,0,centroids[:][i][0],0], ndmin=2)
			else:
				if i < len(centroids):
					a = np.append(a, [[0,0,centroids[:][i][0],0]], axis=0)
	return a

    def doKmeansMultiCluster(self, state, from_inlambda, to_inlambda):
	# Run kmeans for the measurements of this state and return the centroid point (throughput, latency)
	ctd = defaultdict(dict);
	label = []
	label_viol = []
	if self.memory[state]['arrayMeas'] != None :
		sliced_data = None
		sliced_data_violation = None
		for j in self.memory[state]['arrayMeas']:
			# If this measurement belongs in the slice we're insterested in
			if j[0] >= from_inlambda and j[0] <= to_inlambda:
				# add it
				if self.utils.violation_cluster and j[2] >= float(self.utils.latency_threshold):
					if sliced_data_violation == None:
						sliced_data_violation = np.array(j, ndmin=2)
					else:
						sliced_data_violation = np.append(sliced_data_violation, [j], axis=0)
				else:
					if sliced_data == None:
						sliced_data = np.array(j, ndmin=2)
					else:
						sliced_data = np.append(sliced_data, [j], axis=0)
				

		k = int(self.utils.num_of_clusters)
		# 1. No known lamdba values close to current lambda measurement
		if (sliced_data != None): 
			if (len(sliced_data) < k):
#				print "Not enough meas close to current lambda measurement. Producing extra artificial values!";
				thr, lat = self.utils.interpolate_metrics(self.memory,state,from_inlambda)
				#print state
				#print from_inlambda
				if(thr != thr):
					thr = self.load_generator.thr(str(state),from_inlambda)
				if(lat != lat):
					lat = self.load_generator.lat(str(state),from_inlambda)
				for i in range(len(sliced_data),k):
					self.memory[state]['arrayMeas'] = np.append(self.memory[state]['arrayMeas'],[[float(from_inlambda), thr, lat, float(100)]], axis=0)
				for j in self.memory[state]['arrayMeas']:
					#self.my_logger.debug("DOKMEANS self.memory[state]['arrayMeas'][j]: "+ str(j))
					# If this measurement belongs in the slice we're insterested in
					if j[0] >= from_inlambda and j[0] <= to_inlambda:
						if self.utils.violation_cluster and j[2] >= self.utils.latency_threshold:
							if sliced_data_violation == None:
								sliced_data_violation = np.array(j, ndmin=2)
							else:
								sliced_data_violation = np.append(sliced_data_violation, [j], axis=0)
						else:
							if sliced_data == None:
								sliced_data = np.array(j, ndmin=2)
							else:
								sliced_data = np.append(sliced_data, [j], axis=0)
				self.my_logger.debug("DOKMEANS length of sliced_data to be fed to kmeans: "+ str(len(sliced_data)))

		if (sliced_data == None and sliced_data_violation == None):  
#			print "No known lamdba values close to current lambda measurement. Returning artificial values!";
			thr, lat = self.utils.interpolate_metrics(self.memory,state,from_inlambda)
			#print state
			#print from_inlambda
			if(thr != thr):
				thr = self.load_generator.thr(str(state),from_inlambda)
			if(lat != lat):
				lat = self.load_generator.lat(str(state),from_inlambda)
			for i in range(0,k):
				self.memory[state]['arrayMeas'] = np.append(self.memory[state]['arrayMeas'],[[float(from_inlambda), thr, lat, float(100)]], axis=0)
			for j in self.memory[state]['arrayMeas']:
				#self.my_logger.debug("DOKMEANS self.memory[state]['arrayMeas'][j]: "+ str(j))
				# If this measurement belongs in the slice we're insterested in
				if j[0] >= from_inlambda and j[0] <= to_inlambda:
					if self.utils.violation_cluster and j[2] >= self.utils.latency_threshold:
						if sliced_data_violation == None:
							sliced_data_violation = np.array(j, ndmin=2)
						else:
							sliced_data_violation = np.append(sliced_data_violation, [j], axis=0)
					else:
						if sliced_data == None:
							sliced_data = np.array(j, ndmin=2)
						else:
							sliced_data = np.append(sliced_data, [j], axis=0)
			self.my_logger.debug("DOKMEANS length of sliced_data to be fed to kmeans: "+ str(len(sliced_data)))
			

		centroids_viol = []
		if sliced_data_violation != None and len(sliced_data_violation) > 0:
#			print "DOKMEANS length of sliced_data_violation to be fed to kmeans: "+ str(len(sliced_data_violation));
			centroids_viol, label_viol = kmeans2(sliced_data_violation, 1, minit='points')
			if sliced_data != None:
				try:
					centroids, label = kmeans2(np.array(sliced_data[:,2],ndmin=2).reshape((-1,1)), k-1, minit='points')
					centroids = self.getNewCentroids(sliced_data,k-1,centroids, label)
				except Exception, e:
					print str(e)
					print str(sliced_data)
			
		else:
#			centroids, label = kmeans2(sliced_data, k, minit='points')
			if(len(sliced_data) < k):
#				print "sliced_data size smaller than k: " + str(len(sliced_data))
				centroids, label = kmeans2(np.array(sliced_data[:,2],ndmin=2).reshape((-1,1)), len(sliced_data), minit='points')
#				for i in range(0,k-len(sliced_data)):
#					centroids = np.append(centroids,[[0]],axis=0)
#					label = np.append(label,[0],axis=0)
			else:
				centroids, label = kmeans2(np.array(sliced_data[:,2],ndmin=2).reshape((-1,1)), k, minit='points')
			centroids = self.getNewCentroids(sliced_data,k,centroids, label)


		num_of_meas = {}
		#num_of_meas = {'0': 0, '1': 0, '2': 0, '3': 0, '4': 0}
		for j in range(0, k):
			num_of_meas[str(j)] = 0
		meas_sum = 0;
		if len(label) > 0:
			for i in label:
				num_of_meas[str(i)] += 1 
				meas_sum += 1;
		
		num_of_meas_viol = {}
		#num_of_meas = {'0': 0, '1': 0, '2': 0, '3': 0, '4': 0}
		for j in range(0, 1):
			num_of_meas_viol[str(j)] = 0
		if len(label_viol) > 0:
			for i in label_viol:
				num_of_meas_viol[str(i)] += 1 
				meas_sum += 1;


		if not self.utils.multicluster:
			max_meas_cluster = max(num_of_meas.iteritems(), key=operator.itemgetter(1))[0]
			if len(label) > 0:
				for j in label:
					ctd[int(j)]['cluster_weight'] = 1.0 if int(j) == int(max_meas_cluster) else 0;
					ctd[int(j)]['inlambda'] = centroids[int(j)][0];
					ctd[int(j)]['throughput'] = centroids[int(j)][1];
					ctd[int(j)]['latency'] = centroids[int(j)][2];
					ctd[int(j)]['cpu'] = centroids[int(j)][3];
			else:
				for j in range(0, k):
					ctd[int(j)]['cluster_weight'] = 0.0;
					ctd[int(j)]['inlambda'] = 0.0;
					ctd[int(j)]['throughput'] = 0.0;
					ctd[int(j)]['latency'] = 0.0;
					ctd[int(j)]['cpu'] = 0.0;
		else:
			if len(label) > 0:
				for j in label:
					ctd[int(j)]['cluster_weight'] = float(float(num_of_meas[str(j)]) / float(meas_sum));
					ctd[int(j)]['inlambda'] = centroids[int(j)][0];
					ctd[int(j)]['throughput'] = centroids[int(j)][1];
					ctd[int(j)]['latency'] = centroids[int(j)][2];
					ctd[int(j)]['cpu'] = centroids[int(j)][3];
			else:
				for j in range(0, k):
					ctd[int(j)]['cluster_weight'] = 0.0;
					ctd[int(j)]['inlambda'] = 0.0;
					ctd[int(j)]['throughput'] = 0.0;
					ctd[int(j)]['latency'] = 0.0;
					ctd[int(j)]['cpu'] = 0.0;  
   
			if len(label_viol) > 0:
				for j in label_viol:
					ctd[int(self.utils.num_of_clusters)-1]['cluster_weight'] = float(float(num_of_meas_viol[str(0)]) / float(meas_sum));
					ctd[int(self.utils.num_of_clusters)-1]['inlambda'] = centroids_viol[int(j)][0];
					ctd[int(self.utils.num_of_clusters)-1]['throughput'] = centroids_viol[int(j)][1];
					ctd[int(self.utils.num_of_clusters)-1]['latency'] = centroids_viol[int(j)][2];
					ctd[int(self.utils.num_of_clusters)-1]['cpu'] = centroids_viol[int(j)][3];     

	else:
		print "DOKMEANS self.memory[state]['arrayMeas'] is None :|";

	return ctd;

    def moving_average(self, iterable, n=3):
        # moving_average([40, 30, 50, 46, 39, 44]) --> 40.0 42.0 45.0 43.0
        # http://en.wikipedia.org/wiki/Moving_average
        it = iter(iterable)
        d = deque(itertools.islice(it, n-1))
        d.appendleft(0)
        s = sum(d)
        for elem in it:
            s += elem - d.popleft()
            d.append(elem)
            yield s / float(n)
            
    def failuresCalculator(self,currentState, nextState):
    	successfullActions = 0
    	if currentState > nextState: #remove action
    		removeVmsNum = currentState - nextState
		pn = self.utils.removeFailureProb*pow(self.utils.failure_alpha_const,removeVmsNum)*pow(self.utils.failure_beta_const,self.generate_n_current());
    		for i in range(0,removeVmsNum):
    			successfullActions = successfullActions + 1 - self.generate(pn)
    		self.experiment_log.log_failure(str(currentState)+','+str(nextState)+",remove_"+str(removeVmsNum)+','+str(successfullActions)+','+str(removeVmsNum-successfullActions)+','+str(len(self.actionsList))+','+str(self.possibleActionCount)+','+str(self.timeout))
    		return currentState - successfullActions
    	else: #add action
    		addVmsNum = nextState - currentState
		pn = self.utils.addFailureProb*pow(self.utils.failure_alpha_const,addVmsNum)*pow(self.utils.failure_beta_const,self.generate_n_current());
    		for i in range(0,addVmsNum):
    			successfullActions = successfullActions + 1 - self.generate(pn)
    		self.experiment_log.log_failure(str(currentState)+','+str(nextState)+",add_"+str(addVmsNum)+','+str(successfullActions)+','+str(addVmsNum-successfullActions)+','+str(len(self.actionsList))+','+str(self.possibleActionCount)+','+str(self.timeout))
    		return currentState + successfullActions
    	
    def generate(self,p):
    	return int(random.random() < p)
    		
    def generate_n_current(self):
	return self.n_current

    def getMetricConsts(self, from_inlambda, to_inlambda, p, allmetrics):
	latencyConst = defaultdict(dict);
	throughputConst = defaultdict(dict);
	cluster_weights = defaultdict(dict);
	for i in range(int(self.utils.min_cluster_size), int(self.utils.max_cluster_size)+1):
	    vals = defaultdict(dict);          
	    met = defaultdict(dict);
	    if self.measurementsPolicy.startswith('average'):
		met = self.getAverages(str(i))
	    elif self.measurementsPolicy.startswith('centroid'):
		met[i] = self.doKmeansMultiCluster(str(i), from_inlambda, to_inlambda);
		for j in range(0,int(self.utils.num_of_clusters)):
	    		vals[j]['num_nodes'] = i
			if met[i][j] != None and len(met[i][j]) > 0 :
				vals[j]['inlambda'] = met[i][j]['inlambda']
				vals[j]['throughput'] = met[i][j]['throughput']
				vals[j]['latency'] = met[i][j]['latency']
				vals[j]['cpu'] = met[i][j]['cpu']
			else:
				# No clue for this state use defaults/current measurements...
				vals[j]['inlambda'] = allmetrics['inlambda']
				vals[j]['throughput'] = allmetrics['throughput']
				vals[j]['latency'] = allmetrics['latency']

			if int(self.currentState) == i and p == 0:
				vals[j]['inlambda'] = allmetrics['inlambda']
				vals[j]['latency'] = allmetrics['latency']
				vals[j]['throughput'] = allmetrics['throughput']
#			print "CURRENT LATENCY:" + str(allmetrics['latency'])
			allmetrics['num_nodes'] = i
			allmetrics['max_num_nodes'] = int(self.utils.max_cluster_size)
			total_cost,norm_total_cost = self.get_the_cost()
		    	allmetrics['total_cost'] = total_cost
			allmetrics['norm_total_cost'] = norm_total_cost
			allmetrics['norm_latency'] = self.get_norm_latency(allmetrics['latency'],np.mean(self.latencies),np.std(self.latencies))
			if(float(float(allmetrics['latency'])) > float(self.utils.latency_threshold)):
				cur_gain = float(self.utils.gain_punishment);
			else:
			    	cur_gain = float(eval(self.utils.gain,allmetrics))
			self.memory[str(i)]['r'] = cur_gain
	
			if(len(latencyConst[i]) == 0):
				latencyConst[i] = defaultdict(dict);
	
			if(len(throughputConst[i]) == 0):
				throughputConst[i] = defaultdict(dict);
	
			if(len(cluster_weights[i]) == 0):
				cluster_weights[i] = defaultdict(dict);


			if (vals[j]['latency'] == None):
				latencyConst[i][j] = 0;
			else:
				latencyConst[i][j] = vals[j]['latency'];


			if (vals[j]['throughput'] == None):
				throughputConst[i][j] = 0;
			else:
				throughputConst[i][j] = vals[j]['throughput'];


			#Adding cluster_weight = 1 for the first cluster of the current state and 0 sfor the other cluster		
			if int(self.currentState) == i and p == 0:
				if j == 0:
					cluster_weights[i][j] = 1.0;
				else:
					cluster_weights[i][j] = 0.0;
			else:
				if (len(met[i][j]) == 0):
					cluster_weights[i][j] = 0;
				else:
					cluster_weights[i][j] = met[i][j]['cluster_weight'];
	
	return latencyConst,throughputConst, cluster_weights
    	
    def predict_lambda(self,steps_ahead,d_l):
	#*10 as want 5 minutes steps and the load is in 30seconds intervals
	load = self.load_generator.get_next_load(steps_ahead*10) + self.load_generator.myR(-steps_ahead*100, steps_ahead*100);
	self.my_logger.debug("Predicted Lambda step: " + str(steps_ahead) + " -> " + str(load))
	return load;
	
#	self.my_logger.debug("Current Lambda: " + str(d_l))
#	today = datetime.datetime.now()
#	data_index = pandas.date_range(end=today, periods = len(self.past_lambda),freq='30S')
#	data_series = pandas.Series(self.past_lambda, index = data_index)
#	try:
#		model = sm.tsa.ARMA(data_series,(2,1)).fit()
#	except Exception,e: 
#		self.my_logger.debug("An Exception Occured on ARMA")

#	data_series = pandas.Series(self.past_lambda, index = data_index)[len(self.past_lambda)-10:]
#	X = np.array(range(1,len(data_series.values)+1))
#	X = X.reshape((-1,1))
#	y = np.array(data_series.values)
#	model = LinearRegression()
#	model.fit(X, y)
#	inlambda = model.predict([[len(data_series.values)+steps_ahead]])[0]
#	if(inlambda < 0):
#		inlambda = float(d_l)
#	self.my_logger.debug("Linear Regression: " + str(inlambda))
#	return inlambda
	#Big Brother is watching you!
	#prediction = model.predict(str(today), str(today+datetime.timedelta(seconds=steps_ahead*30)), dynamic=True)
	#self.my_logger.debug("Predicted Lambda: " + str(prediction[len(prediction)-1]))
#	return prediction[len(prediction)-1]

    
    def get_norm_latency(self,latency, mean_lat, std_lat):
	z_latency = 0 
	if((latency-mean_lat)/std_lat>1):
		z_latency = 1
	elif((latency-mean_lat)/std_lat<(-1)):
		z_latency = -1 
	else: 
		z_latency = (latency-mean_lat)/std_lat;
	return (z_latency+1.0)/2.0;

	
    def get_the_cost(self):
	total_cost = 0
	for vm_type in self.vm_types_currentState:
		if('_sp' not in vm_type):
			total_cost = total_cost + self.vm_types_currentState[vm_type]*self.utils.vm_types_hour_cost[vm_type]
		else:
			total_cost += self.vm_types_currentState[vm_type]*self.spotHandler.getMarketPrice(self.simTime, self.utils.spotInstanceType)

	min_total_cost = 0
	for vm_type in self.utils.vm_types_hour_cost.keys():
		min_total_cost = min_total_cost + self.utils.vm_types_hour_cost[vm_type]*self.utils.vm_types_min[vm_type]
	
	max_total_cost = 0
	for vm_type in self.utils.vm_types_hour_cost.keys():
		max_total_cost = max_total_cost + self.utils.vm_types_hour_cost[vm_type]*self.utils.vm_types_max[vm_type]

	norm_total_cost = (total_cost-min_total_cost)/(max_total_cost-min_total_cost)
	return total_cost,norm_total_cost

    def ewma(self, alist):
	today = datetime.datetime.now()
	data_index = pandas.date_range(end=today, periods = len(alist),freq='30S')
	data = np.array(alist)
	data_series = pandas.Series(data, index = data_index)

	a = 0.85
	z = pandas.ewma(data_series,com=a/(1.0-a))
	return z.values[len(z.values)-1]

    def takeDecision(self, rcvallmetrics, load_generator, simDatetime, simTime):
        '''
         this method reads allmetrics object created by MonitorVms and decides to change the number of participating
         virtual nodes.
        '''
	self.simTime = simTime
	if(self.utils.spot):
		current_market_price = self.spotHandler.getMarketPrice(self.simTime, self.utils.spotInstanceType);
		self.my_logger.debug("Current Market Price: " + str(current_market_price) + " simTime: " + str(self.simTime))
		for vm_type in self.utils.vm_types_hour_cost:
			if("_sp" in vm_type):
				self.utils.vm_types_hour_cost[vm_type] = current_market_price


	if(len(self.spotHandler.fetchSpots()) > 0):
		costRemoveTuple = self.spotHandler.checkSpots(self.simTime)
		if(len(costRemoveTuple[1]) != 0):
			self.my_logger.debug("Forced to remove: " + str(len(costRemoveTuple[1])) + " spots")
			self.cost_before_remove = self.cost_before_remove + costRemoveTuple[0]
			for id in costRemoveTuple[1]:
				con = sqlite.connect(self.utils.db_file)
				cur = con.cursor()
				try:
				    self.nextState = str(int(self.nextState)-1)
				    vm_type = cur.execute('select vm_type from instances where id='+str(id)).fetchone()[0]
				    self.vm_types_nextState[vm_type] = self.vm_types_nextState[vm_type]-1;
				    #self.currentState = str(self.nextState)
				    #self.vm_types_currentState = self.vm_types_nextState
				    self.countdown = 0
				    cur.execute('delete from instances where id='+str(id))
				    con.commit()
				except sqlite.DatabaseError:
				    print "ERROR in delete"
				cur.close()
				con.close()

	self.load_generator = load_generator;
	#This simulates the delay to add or remove VMs
	if self.countdown > 1:
		self.countdown = self.countdown - 1
	else:
		if len(self.acted) > 1:
			self.acted.pop()
		self.countdown = 0
		self.currentState = str(self.nextState)
		self.vm_types_currentState = self.vm_types_nextState

	startTime = datetime.datetime.now();
	#self.my_logger.debug("start_on: " + str(startTime));

        ## Take decision based on metrics
        action = "none"
        
#        self.my_logger.debug("TAKEDECISION rcvallmetrics: " + str(rcvallmetrics))
        allmetrics = None
        allmetrics = rcvallmetrics.copy()
        self.my_logger.debug("TAKEDECISION state: " + str(self.currentState))
        
        if not allmetrics.has_key('inlambda'):
            allmetrics['inlambda'] = 0
            
        if not allmetrics.has_key('throughput'):
            allmetrics['throughput'] = 0
            
        if not allmetrics.has_key('latency'):
            allmetrics['latency'] = 0
            
        if not allmetrics.has_key('cpu'):
            allmetrics['cpu'] = 0
        
        if not self.debug :
            ## Aggreggation of YCSB client metrics
            clients = 0
            nodes = 0

        if not self.acted[len(self.acted)-1].startswith("done") :
            # Take decision as soon as decisions are unblocked.
            self.waitForIt = 0
            self.my_logger.debug("Last action "+ str(self.acted[len(self.acted)-1]) +" hasn't finished yet, see you later!")
            if self.acted[len(self.acted)-1].startswith("remove"):
                #1. Save the current metrics in memory.
                self.addMeasurement([str(self.currentState), allmetrics['inlambda'], allmetrics['throughput'], allmetrics['latency'], 
                             allmetrics['cpu'], datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")], True)
            elif self.acted[len(self.acted)-1].startswith("add"):
                self.my_logger.debug("Discarding measurement: "+ str([str(self.currentState), allmetrics['inlambda'], allmetrics['throughput'], allmetrics['latency'], 
                             allmetrics['cpu'], datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")]))
                dms = open(self.utils.install_dir+'/'+self.utils.logs_folder_name+'/discarded-measurements.txt', 'a')
                dms.write(str(self.currentState)+'\t\t'+str(allmetrics['inlambda'])+'\t\t'+str(allmetrics['throughput'])+'\t\t'+
                     str(allmetrics['latency'])+'\t\t'+str(allmetrics['cpu'])+'\t\t'+
                     datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")+'\n')
                dms.close()

		meas_dict =  {'num_nodes': float(self.currentState),'inlambda': allmetrics['inlambda'],'throughput': allmetrics['throughput'],'latency': allmetrics['latency'],'cpu': allmetrics['cpu']}       
		meas_dict['max_num_nodes'] = int(self.utils.max_cluster_size)
	        total_cost,norm_total_cost = self.get_the_cost()
    	        meas_dict['total_cost'] = total_cost
	        meas_dict['norm_total_cost'] = norm_total_cost
	        meas_dict['norm_latency'] = self.get_norm_latency(meas_dict['latency'],np.mean(self.latencies),np.std(self.latencies))		
		if(float(float(meas_dict['latency'])) > float(self.utils.latency_threshold)):
			gain = float(self.utils.gain_punishment);
		else:
			gain = float(eval(self.utils.gain,meas_dict))
		transition_cost = float(eval(self.utils.trans_cost,meas_dict))
 
	        not_attack_probability = 1;
	        for vm_type in self.vm_types_currentState:
		    not_attack_probability = not_attack_probability * pow(1-self.utils.vm_types_attack_probs[vm_type],self.vm_types_currentState[vm_type])
   		attack_probability = 1 - not_attack_probability
		attack_incident = self.generate(attack_probability)                
		self.experiment_log.log_discarded_measurement(str(self.currentState)+','+str(allmetrics['inlambda'])+','+str(allmetrics['throughput'])+','+str(allmetrics['latency'])+','+str(allmetrics['cpu'])+','+str(gain-transition_cost)+','+str(attack_incident))
		if eval(self.utils.window_inlambda):
			self.latestMeasQueue = deque([])
            return True
        
        #1. Save the current metrics in memory.
        self.addMeasurement([str(self.currentState), allmetrics['inlambda'], allmetrics['throughput'], allmetrics['latency'], 
                             allmetrics['cpu'], datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")], True)
        
        if self.waitForIt == 0:
            self.waitForIt = 10
        else:
            self.my_logger.debug("New decision in " + str(float(self.waitForIt)/2) + " mins, see you later!")
            self.waitForIt = self.waitForIt - 1            
            return True
        
        
        dif = 0
        from_inlambda = 0.95 * allmetrics['inlambda']
        to_inlambda = 1.05 * allmetrics['inlambda']

	if eval(self.utils.window_inlambda):
		if len(self.latestMeasQueue) > 0:
			cnt = 0;
			inlambda_sum = 0;
			throughput_sum = 0;
			latency_sum = 0;
			for j in self.latestMeasQueue:
				self.my_logger.debug("cnt: " + str(cnt) + " load: " + str(j[0]))
				self.my_logger.debug("cnt: " + str(cnt) + " thr: " + str(j[1]))
				self.my_logger.debug("cnt: " + str(cnt) + " lat: " + str(j[2]))
				inlambda_sum = inlambda_sum+float(j[0])
				throughput_sum = throughput_sum + float(j[1])
				latency_sum = latency_sum + float(j[2])
				cnt = cnt+1
				
			allmetrics['inlambda'] = inlambda_sum/cnt
			allmetrics['throughput'] = throughput_sum/cnt
			allmetrics['latency'] = latency_sum/cnt
			from_inlambda = 0.95 * allmetrics['inlambda']
			to_inlambda = 1.05 * allmetrics['inlambda']
	elif eval(self.utils.ewma_inlambda):
		if len(self.lambdaQueue) > 0: 
			allmetrics['inlambda'] = self.ewma(self.lambdaQueue)
			allmetrics['throughput'] = self.ewma(self.thrQueue)
			allmetrics['latency'] = self.ewma(self.latQueue)

        self.my_logger.debug("TAKEDECISION current lambda: " + str(allmetrics['inlambda']) + " lambda range: " + 
                             str(from_inlambda) + " - " + str(to_inlambda))
        # too low to care, the initial num of nodes can answer 1000 req/sec, so consider it as 0 1000 * len(cluster.size)!!
        if 0.0 < to_inlambda < 1000:
            from_inlambda = 0.0
            #to_inlambda = 0.0
            self.my_logger.debug("TAKEDECISION current lambda: " + str(allmetrics['inlambda']) + " changed lambda range to: " + 
                             str(from_inlambda) + " - " + str(to_inlambda))
	
	predicted_lambdas = {}
	predicted_lambdas[0] = [allmetrics['inlambda']]
	self.my_logger.debug("Current Lambda: " + str(allmetrics['inlambda']))
	for p in range(1,int(self.utils.max_steps)+1):
		predicted_lambdas[p] = []
		predicted_lambdas[p].append(self.predict_lambda(p,allmetrics['inlambda']))
	#initialization	
	latencyConstDict = {0:[]}
	throughputConstDict = {0:[]}
	cluster_weightsDict = {0:[]}
	 
	for p in range(0,int(self.utils.max_steps)+1):
		for a in range(0,len(predicted_lambdas[p])):
			from_inlambda = 0.95 * predicted_lambdas[p][a]
			to_inlambda = 1.05 * predicted_lambdas[p][a]
			if 0.0 < to_inlambda < 1000:
		    		from_inlambda = 0.0

			latencyConst, throughputConst, cluster_weights = self.getMetricConsts(from_inlambda, to_inlambda, p, allmetrics)
			latencyConstDict[a].append(latencyConst)
			throughputConstDict[a].append(throughputConst)
			cluster_weightsDict[a].append(cluster_weights)

	vm_types_num_nodes={}
	for vm_type in self.vm_types_currentState:
		vm_types_num_nodes[vm_type] = [self.utils.vm_types_min[vm_type],self.utils.vm_types_max[vm_type],self.vm_types_currentState[vm_type]]


	cost_spot_step = {}
	for vm_type in self.vm_types_currentState:
		if("_sp" in vm_type):
			cost_spot_step[vm_type] = [self.utils.vm_types_hour_cost[vm_type]] * (int(self.utils.max_steps)+1)
			if(len(self.spotHandler.fetchSpots()) > 0):
				for i in range(0,int(self.utils.max_steps)+1):
					cost_spot_step[vm_type][i] = self.spotHandler.getMarketPrice(self.simTime+(i*10), self.utils.spotInstanceType)

	spot_bid_price = {}
	for vm_type in self.vm_types_currentState:
		if("_sp" in vm_type):
			spot_bid_price[vm_type] = self.spotHandler.getBidPrice(vm_type, self.utils.spotInstanceType)


	remove_limits_vm_types = {}
	if(self.utils.hourly):
		for vm_type in self.utils.vm_types:
			remove_steps = []
			for i in range(0,int(self.utils.max_steps)+1):
				if i == 0:
					remove_step_0 = min(self.utils.rem_nodes, len(self.utils.get_remove_candidate_instances_sim(simDatetime,self.utils.last_vm_secs,vm_type)))
					remove_steps.append(remove_step_0)
				else:
					#how many can I remove after i*5mins later?
					remove_step_x = min(self.utils.rem_nodes, len(self.utils.get_remove_candidate_instances_sim(simDatetime+datetime.timedelta(0,i*300),self.utils.last_vm_secs,vm_type)))
					remove_steps.append(remove_step_x)
			remove_limits_vm_types[vm_type] = remove_steps


	self.actionsList = []
	#self.my_logger.debug("TAKING DECISION USING PRISM!");
	mean_lat = np.mean(self.latencies)
	std_lat = np.std(self.latencies)
	if self.utils.mat:
		prismAction = self.mdp_handler.getAction(self.currentState, latencyConst, throughputConst, cluster_weights)[0];
	else:
		prismResultAction = self.mdp_handler.getAction(self.currentState, latencyConstDict, throughputConstDict, cluster_weightsDict, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, mean_lat=mean_lat, std_lat=std_lat);
		prismAction = prismResultAction[0]
		self.actionsList = prismResultAction[1]
		self.possibleActionCount = prismResultAction[2]
		self.timeout = prismResultAction[3]
#		print 'prismResultAction: ' + str(prismResultAction)
	

	total_cost = 0
	for vm_type in self.vm_types_currentState:
		if('_sp' not in vm_type):
			total_cost = total_cost + self.vm_types_currentState[vm_type]*self.utils.vm_types_hour_cost[vm_type]
		else:
			total_cost += self.vm_types_currentState[vm_type]*self.spotHandler.getMarketPrice(self.simTime, self.utils.spotInstanceType)

	allmetrics['total_cost'] = total_cost
	allmetrics['num_nodes'] = int(self.currentState)
	if(prismAction.startswith("add")):
		self.countdown = 10*self.utils.transient_counter_limit
#		This goes with the first lines of this method to simulate the delay of the VMs addition
		self.acted.append('add')
		prism_add = int(prismAction.split("_")[1]);
		add_limit = int(self.utils.add_nodes);
		self.nextState = int(self.currentState) + ( add_limit if prism_add>add_limit else prism_add);

		con = sqlite.connect(self.utils.db_file)
		cur = con.cursor()
		for vm_type in self.vm_types_nextState:
			self.vm_types_nextState[vm_type] = self.vm_types_nextState[vm_type]+prismAction.count(vm_type)

			startID = int(cur.execute('select max(id) from instances').fetchone()[0])+1
			for i in range(startID, startID+prismAction.count(vm_type)):
				self.experiment_log.log_cost("Actual time: "+str(simDatetime)+" - Adding "+vm_type+" node: "+str(i))
				
				
				cur.execute(""" insert into instances(id, launch_time, vm_type)
						    values  (?,?,?)""",(i, str(time.mktime(simDatetime.timetuple())),vm_type))
				con.commit()


				if('_sp' in vm_type):
					self.spotHandler.provisionSpot(self.utils.spotInstanceType,1,self.simTime,i,vm_type)
				
		cur.close()
		con.close()

		action = prismAction
	elif(prismAction.startswith("rem")):
		self.countdown = 10
		self.acted.append('remove')
		prism_rem = int(prismAction.split("_")[1]);
		rem_limit = int(self.utils.rem_nodes);
		self.nextState = int(self.currentState) - (prism_rem if prism_rem<rem_limit else rem_limit);
		for vm_type in self.vm_types_nextState:
			self.vm_types_nextState[vm_type] = self.vm_types_nextState[vm_type]-prismAction.count(vm_type)

		if(self.utils.multiType):
			for vm_type in self.vm_types_nextState:
				to_remove = []
				if(self.utils.hourly):
					vm_type_limit = self.utils.get_remove_candidate_instances_sim(simDatetime,self.utils.last_vm_secs,vm_type)
					to_remove = self.utils.get_remove_candidate_instances_vm_type(vm_type,min(prismAction.count(vm_type),vm_type_limit))
				else:
					to_remove = self.utils.get_remove_candidate_instances_vm_type(vm_type,prismAction.count(vm_type))
				for remove_node in to_remove:
					self.removed_vms_cnt = self.removed_vms_cnt + 1
					con = sqlite.connect(self.utils.db_file)
					cur = con.cursor()
					launch_time = cur.execute('select launch_time from instances where id='+str(remove_node)).fetchone()[0]
					vm_cost = self.utils.get_the_cost(launch_time,time.mktime(simDatetime.timetuple()),vm_type)
					if('_sp' in vm_type):						
						vm_cost = self.spotHandler.stopSpotBasedOnInstanceTableId(self.simTime,remove_node)					
					self.experiment_log.log_cost("Actual time: "+str(simDatetime)+" - Terminating "+vm_type+" node: "+str(remove_node)+" - Running time: "+self.utils.get_running_time(launch_time,time.mktime(simDatetime.timetuple()))+" - Cost: "+str(vm_cost)+" euros")
					self.cost_before_remove = self.cost_before_remove + vm_cost
					
					try:
					    cur.execute('delete from instances where id='+str(remove_node))
					    con.commit()
					except sqlite.DatabaseError:
					    print "ERROR in delete"
					cur.close()
					con.close()
		else:
			for i in range(int(self.nextState), self.utils.get_total_num_vms()):
				self.removed_vms_cnt = self.removed_vms_cnt + 1
				con = sqlite.connect(self.utils.db_file)
				cur = con.cursor()
				launch_time = cur.execute('select launch_time from instances where id='+str(i)).fetchone()[0]
				vm_cost = self.utils.get_the_cost(launch_time,time.mktime(simDatetime.timetuple()))
				self.experiment_log.log_cost("Actual time: "+str(simDatetime)+" - Terminating node: "+str(i)+" - Running time: "+self.utils.get_running_time(launch_time,time.mktime(simDatetime.timetuple()))+" - Cost: "+str(vm_cost)+" euros")
				self.cost_before_remove = self.cost_before_remove + vm_cost
				try:
				    cur.execute('delete from instances where id='+str(i))
				    con.commit()
				except sqlite.DatabaseError:
				    print "ERROR in delete"
				cur.close()
				con.close()
		
		self.experiment_log.log_cost("Actual time: "+str(simDatetime)+" - Paid so far: " + str(self.cost_before_remove) + " euros");
		
		con = sqlite.connect(self.utils.db_file)
		cur = con.cursor()
		cur.execute(""" update exp set removedVMs=?,cost_before_remove=? where id=? """, (self.removed_vms_cnt,self.cost_before_remove,1))
		con.commit()
		cur.close()

		action = prismAction

	endTime = datetime.datetime.now();
	self.my_logger.debug("DECITION TAKEN: " + action);
	allmetrics['num_nodes'] = int(self.currentState)
	allmetrics['max_num_nodes'] = int(self.utils.max_cluster_size)
	
	total_cost,norm_total_cost = self.get_the_cost()
    	allmetrics['total_cost'] = total_cost
	allmetrics['norm_total_cost'] = norm_total_cost
	allmetrics['norm_latency'] = self.get_norm_latency(allmetrics['latency'],np.mean(self.latencies),np.std(self.latencies))

	if(float(float(allmetrics['latency'])) > float(self.utils.latency_threshold)):
		real_gain = float(self.utils.gain_punishment);
	else:
		real_gain = float(eval(self.utils.gain,allmetrics))
	real_transition_cost = float(eval(self.utils.trans_cost, allmetrics))                     
	self.experiment_log.log_action(str(simDatetime)+','+str(endTime-startTime)+','+str(self.currentState)+','+str(self.nextState)+','+str(real_gain-real_transition_cost))

        
        ## ACT
        self.my_logger.debug("Taking decision with acted: " + str(self.acted))
        if self.acted[len(self.acted)-1] == "done" :
                ## start the action as a thread
                #thread.start_new_thread(self.polManager.act, (action, self.acted, self.currentState, self.nextState))
                self.my_logger.debug("Action undertaken: " + str(action) + " current state: " + str(self.currentState)
                                     + " next: " + str(self.nextState))
        else: # shouldn't end up here
                ## Action still takes place so do nothing
                self.my_logger.debug("Waiting for action to finish: " +  str(action) +" "+ str(self.acted))
        
        action = "none"
        
        return True

    
class PolicyManager(object):
    '''
    This class manages and abstracts the policies that Decision Maker uses. 
    '''


    def __init__(self, policyDescription,properties_file):
        '''
        Constructor. Requires a policy description that sets the policy. 
        ''' 
        self.utils = Utils.Utils(properties_file)
        self.pdesc = policyDescription
        ## Install logger
        LOG_FILENAME = self.utils.install_dir+'/'+self.utils.logs_folder_name+'/Coordinator.log'
        self.my_logger = logging.getLogger('PolicyManager')
        self.my_logger.setLevel(self.utils.logging_level)
        
        handler = logging.handlers.RotatingFileHandler(
                      LOG_FILENAME, maxBytes=2*1024*1024, backupCount=5)
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
        handler.setFormatter(formatter)
        self.my_logger.addHandler(handler)
    
    def act (self, action, acted, curr, next):
        self.my_logger.debug("Taking decision with acted: " + str(acted))
        if self.pdesc == "test":
	    if action.startswith("add"):
                num_add = int(next)-int(curr)
                self.my_logger.debug("Launching new instances: " + str(num_add))
                self.my_logger.debug("Launched new instance/instances: ")
                acted.append("add")
                self.my_logger.debug("Running instances: ")
            
                self.my_logger.debug("Nodes added!")
                self.my_logger.debug("The number of nodes has changed, off to reconfigure Ganglia!!")
                self.my_logger.debug("Ganglia configuration is done!! State Changed!")
                self.my_logger.debug("Go to sleep for 5 mins.")
                #self.my_logger.debug("Restart master: ")
                #self.NoSQLCluster.restart_master()
                self.my_logger.debug("Unblock decision making!")
                acted.pop()
            if action.startswith("remove"):
                acted.append("remove")
                num_rem = int(curr)-int(next)
                self.my_logger.debug("The number of nodes has changed, off to reconfigure Ganglia!!")
                self.my_logger.debug("Ganglia configuration is done!! State Changed!")
		self.my_logger.debug("Go to sleep for 5 mins.")
                #self.my_logger.debug("Restart master: ")
                #self.NoSQLCluster.restart_master()
                self.my_logger.debug("Unblock decision making!")
                acted.pop() 
#            if not action == "none":
#                self.my_logger.debug("Starting rebalancing for active cluster.")
#                self.NoSQLCluster.rebalance_cluster()
        elif self.pdesc == "debug":
            pass
        action = "none"

if __name__ == '__main__':
	mdp = MDPDecisionMaker_sim()
