from __future__ import unicode_literals

from django.db import models
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class Elasticity(models.Model):
	class Meta:
		verbose_name = u'Elasticity'
		verbose_name_plural = u'Elasticities'

	UTILITY_CHOICES = (
			('r1', u'throughput/num_nodes'),
			('r2', u'1.0/num_nodes'),
			('r3', u'max_num_nodes-num_nodes+1'),
			('r4', u'1.0-(1.0/num_nodes)'),
			('r42', u'2.0-(2.0/(1+norm_total_cost))'),
			('r10', u'0.0*norm_total_cost+1.0*norm_latency'),
			('r9', u'0.1*norm_total_cost+0.9*norm_latency'),
			('r8', u'0.2*norm_total_cost+0.8*norm_latency'),
			('r11', u'0.35*norm_total_cost+0.65*norm_latency'),
			('r7', u'0.3*norm_total_cost+0.7*norm_latency'),
			('r6', u'0.4*norm_total_cost+0.6*norm_latency'),
			('r5', u'0.5*norm_total_cost+0.5*norm_latency'),
			('r12', u'0.6*norm_total_cost+0.4*norm_latency'),
			('r13', u'0.7*norm_total_cost+0.3*norm_latency'),
			('r14', u'0.8*norm_total_cost+0.2*norm_latency'),
			('r15', u'0.9*norm_total_cost+0.1*norm_latency'),
			('r16', u'1.0*norm_total_cost+0.0*norm_latency'),
		    )

	EMULATION_CHOICES = (('okeanos', u'Cassandra on ~okeanos'),)

	DECISION_POLICY_CHOICES = (
			('probbexact', u'ORadv'),
			('noprob', u'OR'),
			('rf', u'BRsimple'),
			('raf', u'BRecon'),
			('probe', u'BR-BRecon'),
			('probb', u'BR-BPstab'),
			('probq', u'BR-BPqual'),
		    )

	SMOOTHING_CHOICES = (
			('wf', u'None'),
			('wt', u'MA'),
			('ewmat', u'EWMA'),
		    )
	TYPES_CHOICES = (
			('CASSANDRA39', u'Apache Cassandra 3.9'),
			('HBASE', u'Apache HBase 0.98')
		    )

	STATE_CHOICES = (
			('running', u'running'),
			('stopped', u'stopped')
		    )

	PREDICITON_CHOICES = (
			('emulated', u'Pseudoprediciton'),
			('lr', u'Linear Regression'),
			('arma', u'ARMA'),
			('arima', u'ARIMA')
		    )


	state = models.CharField(max_length=20, default='stopped', verbose_name="State", choices=STATE_CHOICES)

	cluster_type = models.CharField(max_length=200, default='CASSANDRA39', verbose_name="NoSQL Database", choices=TYPES_CHOICES)
	owner = models.CharField(null=True, max_length=200, verbose_name="DB Owner")
	init_cluster_size = models.IntegerField(default=2, verbose_name="Initial number of NoSQL nodes") 
	
	simulation = models.BooleanField(default=True, verbose_name="Emulate the Deployment")
	min_vms = models.IntegerField(default=2, verbose_name="Min. Cluster Size")
	max_vms = models.IntegerField(default=12, verbose_name="Max. Cluster Size")

	time_steps = models.IntegerField(default=15720, verbose_name="Time Steps (30secs/step)")
	training_dataset = models.CharField(default='okeanos', max_length=300, verbose_name="Training Dataset") 
	emulation_dataset = models.CharField(default='okeanos', max_length=300, choices=EMULATION_CHOICES, verbose_name="Emulation Dataset")
	prism_path = models.CharField(null=True, verbose_name="Prism Model Checker Path", max_length=300)

	utility_function = models.CharField(default='r42', max_length=300, verbose_name="Utility Function", choices=UTILITY_CHOICES)
	latency_threshold = models.IntegerField(default=45, verbose_name="Latency Threshold (ms)")
	punishment = models.FloatField(default=2.0, verbose_name="Punishment Value")
	acceptance_percentage = models.FloatField(default=0.0, verbose_name="Utility Difference [0,1]") 

	decision_policy = models.CharField(max_length=300, choices=DECISION_POLICY_CHOICES, verbose_name="Decision Policy", default="rf")
	smoothing = models.CharField(max_length=300, default="ewmat", choices=SMOOTHING_CHOICES, verbose_name="Smoothing Window")
	hourly = models.BooleanField(default=True, verbose_name="Hourly Charging Consideration")
	
	max_add = models.IntegerField(default=3, verbose_name="Max. Allowed VM Additions")
	max_remove = models.IntegerField(default=2, verbose_name="Max. Allowed VM Removals")
	clusters_num = models.IntegerField(default=3, verbose_name="Num. of Meas. Clusters")
	violation_cluster = models.BooleanField(default=True, verbose_name="Use of Violation Cluster")
	verification_steps = models.IntegerField(default=4, verbose_name="Verification Steps")
	transient_steps = models.IntegerField(default=2, verbose_name="Transient Steps After VM Additions")
	cumulative_reward = models.BooleanField(default=True, verbose_name="Use of Cumulative Rewards")
	attack_model = models.BooleanField(default=False, verbose_name="Consider Security Attacks")

	def __str__(self):
        	return str(self.id)

	
@python_2_unicode_compatible
class NoSQLInstance(models.Model):
	class Meta:
		verbose_name = u'NoSQL Instance'
	elasticity = models.ForeignKey(Elasticity, on_delete=models.CASCADE, verbose_name="Elasticity Handler", null=True)
	image_id = models.CharField(max_length=200, verbose_name="snf-image ID")
	dns_name = models.CharField(max_length=200, verbose_name="DNS Name")
	ip_address = models.CharField(max_length=200, verbose_name="IP Address")
	mac_address = models.CharField(max_length=200, verbose_name="MAC Address")
	vpn = models.CharField(max_length=200, verbose_name="VPN")
	state = models.CharField(max_length=200, verbose_name="State")
	instance_type = models.CharField(max_length=200, verbose_name="Instance Type")
	launch_time = models.DateTimeField(auto_now_add=True, verbose_name="Launch Time")
	placement = models.CharField(max_length=200, verbose_name="Placement")
	username = models.CharField(max_length=200, verbose_name="Username")
	password = models.CharField(max_length=200, verbose_name="Password")

	def __str__(self):
        	return self.ip_address	
