from django.shortcuts import render
from django.http import HttpResponseRedirect, HttpResponse
from django.urls import reverse
from .models import Elasticity
from django.template import loader
from django.shortcuts import get_object_or_404,render
from django.http import Http404
from django.views import generic
from .forms import ElasticityForm
from decision.Experiments import Experiment
import os, signal
from django.conf import settings
import shutil


class IndexView(generic.ListView):
    context_object_name = 'elasticity_list'

    def get_queryset(self):
	qs = Elasticity.objects.values('id','simulation','cluster_type','decision_policy','utility_function','state')
	new_qs = []
	for row in qs:
		row['cluster_type'] = dict(Elasticity.TYPES_CHOICES)[row['cluster_type']]
		row['decision_policy'] = dict(Elasticity.DECISION_POLICY_CHOICES)[row['decision_policy']]
		row['utility_function'] = dict(Elasticity.UTILITY_CHOICES)[row['utility_function']]
		new_qs.append(row)
        return new_qs


def update(request, elasticity_id):
	if request.method == 'POST':
		instance = get_object_or_404(Elasticity, id=elasticity_id) #we need the instance to update
		form = ElasticityForm(request.POST or None, instance=instance)
		form.save()
		return HttpResponseRedirect(reverse('elasticity:elasticity_index'))
	else:
		elasticity = get_object_or_404(Elasticity, id=elasticity_id)
		form = ElasticityForm(instance=elasticity)
		return render(request, 'elasticity/details.html', {
        		'form': form,
			'update': True,
    		})

def create(request):
    if request.method == 'POST':
	form = ElasticityForm(request.POST)
	if(form.is_valid()):	
		form.save()
		return HttpResponseRedirect(reverse('elasticity:elasticity_index'))
	return render(request, 'elasticity/details.html', {
        		'form': form
    		})
    else:
	form = ElasticityForm()
	return render(request, 'elasticity/details.html', {
        	'form': form,
    	})

def delete(request, elasticity_id):
	elasticity = get_object_or_404(Elasticity, pk=elasticity_id)
	elasticity.delete()
	file_path = os.path.join(settings.BASE_DIR, 'decision')+"/exp_"+str(elasticity_id)
	if os.path.exists(file_path):
		shutil.rmtree(file_path)
	return HttpResponseRedirect(reverse('elasticity:elasticity_index'))

def start(request, elasticity_id):
	elasticity = get_object_or_404(Elasticity, pk=elasticity_id)
	install_dir = os.path.join(settings.BASE_DIR, 'decision')
	pid_file = str(elasticity_id)
	iterations = 1

	params = ""
	if(elasticity.simulation):
		params += "sim_mdp2_"
	else:
		params += "real_mdp2_"
	params += elasticity.emulation_dataset.encode('utf-8')+"_"
	params += elasticity.utility_function.encode('utf-8')+"_"
	params += elasticity.smoothing.encode('utf-8')+"_"
	params += str(elasticity.max_add)+"a_"
	params += str(elasticity.max_remove)+"r_"
	params += str(elasticity.init_cluster_size)+"init_"
	params += str(elasticity.min_vms)+"min_"
	params += str(elasticity.max_vms)+"max_"
	params += str(elasticity.latency_threshold)+"l_"
	params += str(elasticity.clusters_num)+"cl_"
	params += str(elasticity.time_steps)+"meas_"
	params += str(elasticity.verification_steps)+"steps_"
	params += str(elasticity.cumulative_reward)+"Cum_"
	params += str(elasticity.violation_cluster)+"Viocl_"
	params += str(elasticity.attack_model)+"Attack_"
	params += str(elasticity.transient_steps)+"trancnt_"
	params += elasticity.decision_policy.encode('utf-8')+"_"
	params += str(elasticity.acceptance_percentage)+"ap_"
	params += str(elasticity.hourly)+"Hourly_"
	params += str(elasticity.punishment)+"pun"

#	params = "sim_mdp2_okeanos_r4_ewmat_3a_2r_8init_8min_18max_50l_3cl_15720meas_4steps_TrueCum_TrueViocl_FalseAttack_2trancnt_rf_0.0ap_TrueHourly_2.0pun"
	exp = Experiment(install_dir,pid_file,iterations,params)
	elasticity.state = "running"
	elasticity.save()
	return HttpResponseRedirect(reverse('elasticity:elasticity_index'))

def stop(request, elasticity_id):
	if(os.path.isfile("/tmp/"+str(elasticity_id))):
		with open("/tmp/"+str(elasticity_id)) as f:
	    		pid = f.readline()
		os.kill(int(pid), signal.SIGKILL)
		os.remove("/tmp/"+str(elasticity_id))
	elasticity = get_object_or_404(Elasticity, pk=elasticity_id)
	elasticity.state = "stopped"
	elasticity.save()
	return HttpResponseRedirect(reverse('elasticity:elasticity_index'))
