from django.shortcuts import render
from django.http import HttpResponseRedirect, HttpResponse
from django.urls import reverse
from django.template import loader
from django.shortcuts import get_object_or_404,render
import json, csv, os, time
from django.http import JsonResponse
from collections import OrderedDict as odict
from elasticity.models import Elasticity
from django.conf import settings
from decision import Utils
from pysqlite2 import dbapi2 as sqlite
from datetime import datetime


def diagrams(request):
	qs = Elasticity.objects.values('id','state')
	return render(request, 'monitoring/diagrams.html',{'elasticity_managers_list' : qs})

def get_the_cost(request, elasticity_id):
	utils = Utils.Utils(get_properties_file(elasticity_id))

	con = sqlite.connect(get_db_file(elasticity_id))
	cur = con.cursor()
	simDate = cur.execute('select simDate from exp where id=1').fetchone()[0]
	con.commit()
	cost_before_remove = cur.execute('select cost_before_remove from exp where id=1').fetchone()[0]
	con.commit()
	cur.close()
	current_exp_time = datetime.strptime(simDate, '%b %d %Y %I:%M:%S%p')
	cost = float(cost_before_remove)+utils.get_the_cost_of_all(current_exp_time)
	return cost

def get_spawnedVMs(request, elasticity_id):
	utils = Utils.Utils(get_properties_file(elasticity_id))
	con = sqlite.connect(get_db_file(elasticity_id))
	cur = con.cursor()
	removedVMs = int(cur.execute('select removedVMs from exp where id=1').fetchone()[0])+utils.get_total_num_vms()
	con.commit()
	cur.close()
	return removedVMs

def get_running_time(request, elasticity_id):
	utils = Utils.Utils(get_properties_file(elasticity_id))
	con = sqlite.connect(get_db_file(elasticity_id))
	cur = con.cursor()
	simDate = cur.execute('select simDate from exp where id=1').fetchone()[0]
	con.commit()
	startDate = cur.execute('select startDate from exp where id=1').fetchone()[0]
	con.commit()
	cur.close()
	running_time = str(utils.get_running_time(time.mktime(datetime.strptime(startDate, '%b %d %Y %I:%M:%S%p').timetuple()), time.mktime(datetime.strptime(simDate, '%b %d %Y %I:%M:%S%p').timetuple())))
	return running_time

def stats(request, elasticity_id):
	lat = get_object_or_404(Elasticity, pk=elasticity_id).latency_threshold


	logs_dir = get_exp_log_dirs(elasticity_id)

	violations = 0.0;
	meas_cnt = 0.0;
	vms_num = 0;

	for root, dirs, files in os.walk(logs_dir):
		for file in files:
			if file == "measurements.log":
				ifile_path = os.path.join(root, file)
				ifile = open(ifile_path,'rb')
				reader = csv.reader(ifile)
				rownum = -1
				for row in reader:
				    	vms_num = vms_num + float(row[3])
					if float(row[6]) > float(lat):
						violations = violations+1
				    	meas_cnt = meas_cnt+1
				ifile.close()
	lat_percent = round(float(violations*100.0/meas_cnt),2)


	instances_stats_list = []
	user_stats_dict = {}
	user_stats_dict['cost'] = round(get_the_cost(request, elasticity_id),2)
	user_stats_dict['spawned_vms'] = get_spawnedVMs(request, elasticity_id)
	user_stats_dict['running_time'] = get_running_time(request, elasticity_id)
	user_stats_dict['violations'] = lat_percent
	instances_stats_list.append(user_stats_dict)
        return_dict = {'aaData': instances_stats_list}
        instances_stats = return_dict
	return JsonResponse(instances_stats)


def stats_lat_vms_load(request, elasticity_id):
	lat_thres = float(get_object_or_404(Elasticity, pk=elasticity_id).latency_threshold)
	logs_dir = get_exp_log_dirs(elasticity_id)
	for root, dirs, files in os.walk(logs_dir):
		for file in files:
			if file == "measurements.log":
				ifile = open(os.path.join(root, file),'rb')
				reader = csv.reader(ifile)
				rownum = -1
				vms = []
				load = []
				lat = []
				LV = {}
				for row in reader:
				    rownum = rownum+1
				    if(rownum == 15720):
					break
				    vms.append(int(row[3]))
				    load.append(float(row[4]))
				    lat.append(float(row[6]))
				ifile.close()
				categories = range(0,rownum)
				return JsonResponse( {'categories':categories, 'vms':vms, 'load':load, 'lat':lat, 'lat_thres':lat_thres, 'vms_title':'VMs Number Adaptation to Incoming Load', 'lat_title':'Response Latency Variation to Incoming Load', 'subtitle':'Elton Statistics ;-)'})

	return JsonResponse({})


def get_exp_log_dirs(elasticity_id):
	emulation = bool(get_object_or_404(Elasticity, pk=elasticity_id).simulation)
	exp_logs_prefix = 'sim_'
	if(not emulation):
		exp_logs_prefix = 'real_'
	return os.path.join(settings.BASE_DIR, 'decision')+"/exp_"+str(elasticity_id)+"/"+exp_logs_prefix+str(elasticity_id)


def get_properties_file(elasticity_id):
	emulation = bool(get_object_or_404(Elasticity, pk=elasticity_id).simulation)
	exp_logs_prefix = 'sim_'
	if(not emulation):
		exp_logs_prefix = 'real_'
	return os.path.join(settings.BASE_DIR, 'decision')+"/exp_"+str(elasticity_id)+"/Coordinator_"+str(elasticity_id)+".properties"


def get_db_file(elasticity_id):
	emulation = bool(get_object_or_404(Elasticity, pk=elasticity_id).simulation)
	exp_logs_prefix = 'sim_'
	if(not emulation):
		exp_logs_prefix = 'real_'
	return os.path.join(settings.BASE_DIR, 'decision')+"/exp_"+str(elasticity_id)+"/eMRE.logs_"+str(elasticity_id)+".tiramola.db"
