#!/usr/bin/env python

import Utils,re,operator,logging, subprocess, time, math, sys;
from ModelHandler import ModelHandler;
from os import remove;

class PrismHandler(): 

	def __init__(self, properties_file):
		self.utils = Utils.Utils(properties_file)
		self.modelFilePath = self.utils.modelFilePath
		self.prismFilePath = self.utils.prismFilePath
		self.prismSMGFilePath = self.utils.prismSMGFilePath
		self.propertiesFilePath = self.utils.propertiesFilePath
		self.propertiesProbabilisticFilePath = self.utils.propertiesProbabilisticFilePath
		self.propertiesSMGFilePath = self.utils.propertiesSMGFilePath
		self.initNumVms = self.utils.initial_cluster_size
		self.minNumVms = self.utils.min_cluster_size
		self.maxNumVms = self.utils.max_cluster_size
		self.num_of_clusters = self.utils.num_of_clusters
		self.mh = ModelHandler(properties_file)
		self.possibleActionCount = 1
		self.timeout = 0
		self.pf = None


	def verifyPCTL(self, pctlProperty, init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes={}, remove_limits_vm_types = {}, cost_spot_step={}, spot_bid_price={}, smg=False, probabilistic=False, allowed_actions=[],mean_lat=-1,std_lat=-1):
		self.pf = pctlProperty
		return self.callPrism(init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, allowed_actions=allowed_actions, mean_lat=mean_lat, std_lat=std_lat)

	def getAction(self, init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, mean_lat=-1, std_lat=-1):
		#self.mean_lat = mean_lat
		#self.std_lat = std_lat

		if self.utils.multiType:
			return (self.getActionMultiType(init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price,mean_lat=mean_lat,std_lat=std_lat),[],self.possibleActionCount,self.timeout)


	def getActionMultiType(self, init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price,mean_lat=-1,std_lat=-1):
		#print "calling prism";
		out = self.callPrism(init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, allowed_actions=["add","remove","no_op"],mean_lat=mean_lat,std_lat=std_lat);
#		print out;
		sys.stdout.flush()

		action,reward = self.parsePrismOutput(out,init_num_nodes,vm_types_num_nodes)
		if('no_action' in action):
			return "no_op_"+str(reward) 
		#accepted_percentage
		a_p = self.utils.accepted_percentage
		if("add" in action and self.utils.removeFirstPolicy):
			out = self.callPrism(init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, allowed_actions=["remove","no_op"],mean_lat=mean_lat,std_lat=std_lat);
			rem_action,rem_reward = self.parsePrismOutput(out,init_num_nodes,vm_types_num_nodes)
			if(rem_action != 'no_action' and float(rem_reward) <= float(reward)+(a_p*float(reward))):
#				print "RemoveFirst policy reward is +- "+str(a_p)+" of the standard reward"
#				print "RemoveFirst policy reward: " + str(rem_reward)
#				print "Standard reward: " + str(reward)
#				print "Standard Action: " + action
#				print "RemoveFirst Action: " + rem_action
				return rem_action+"_"+str(rem_reward)
			else:
				return action+"_"+str(reward)
		elif("add" in action and self.utils.removeFirstAllCasesPolicy):
			possible_actions = {}
			for i in range(1,self.utils.rem_nodes+1):
				if(int(init_num_nodes)-i < int(self.utils.min_cluster_size)):
					continue
				out = self.callPrism(init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, allowed_actions=["rem_"+str(i)],mean_lat=mean_lat,std_lat=std_lat);
				rem_action,rem_reward = self.parsePrismOutput(out,init_num_nodes,vm_types_num_nodes)
				if(rem_action != 'no_action' and float(rem_reward) <= float(reward)+(a_p*float(reward)) ):
					possible_actions[rem_action] = rem_reward
			out = self.callPrism(init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, allowed_actions=["no_op"],mean_lat=mean_lat,std_lat=std_lat);
			no_op_action,no_op_reward = self.parsePrismOutput(out,init_num_nodes,vm_types_num_nodes)
			if(no_op_action != 'no_action' and float(no_op_reward) <= float(reward)+(a_p*float(reward)) ):
				possible_actions[no_op_action] = no_op_reward

			if(len(possible_actions) == 0):
				return action+"_"+str(reward)
			elif(len(possible_actions) == 1):
				rem_action = possible_actions.keys()[0]
				rem_reward = possible_actions[possible_actions.keys()[0]]
#				print "RemoveFirstAllCases policy reward is +- "+str(a_p)+" of the standard reward"
#				print "RemoveFirstAllCases policy reward: " + str(rem_reward)
#				print "Standard reward: " + str(reward)
#				print "Standard Action: " + action
#				print "RemoveFirst Action: " + rem_action
				return rem_action+"_"+str(rem_reward)
			else:
				rem_action = sorted(possible_actions.keys(),reverse=True)[0]
				rem_reward = possible_actions[rem_action]
#				print "RemoveFirstAllCases policy rewards are +- "+str(a_p)+" of the standard reward"
#				print "Found Actions: " + str(len(possible_actions))
#				print "RemoveFirstAllCases policy reward: " + str(rem_reward)
#				print "Standard reward: " + str(reward)
#				print "Standard Action: " + action
#				print "RemoveFirst Action: " + rem_action
				return rem_action+"_"+str(rem_reward)
		elif(self.utils.realProbabilisticPolicy):
			prob_policy = self.utils.prob_policy
			if (prob_policy == 'economy' and "add" not in action and a_p != 0):
				return action+"_"+str(reward)
			out = open(self.propertiesProbabilisticFilePath, 'w');
			out.write('Pmax=? [ F (phase=4)&(latency>latency_threshold)]\n');
			out.close();

			possible_actions = {}
			for i in range(1,self.utils.add_nodes+1):
				if(int(init_num_nodes)+i > int(self.utils.max_cluster_size)):
					continue
				out = self.callPrism(init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, allowed_actions=["add_"+str(i)],mean_lat=mean_lat,std_lat=std_lat);
				add_action,add_reward = self.parsePrismOutput(out,init_num_nodes,vm_types_num_nodes)
				if(add_action != 'no_action' and float(add_reward) <= float(reward)+(a_p*float(reward)) ):
					possible_actions[add_action] = add_reward
			for i in range(1,self.utils.rem_nodes+1):
				if(int(init_num_nodes)-i < int(self.utils.min_cluster_size)):
					continue
				out = self.callPrism(init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, allowed_actions=["rem_"+str(i)],mean_lat=mean_lat,std_lat=std_lat);
				rem_action,rem_reward = self.parsePrismOutput(out,init_num_nodes,vm_types_num_nodes)
				if(rem_action != 'no_action' and float(rem_reward) <= float(reward)+(a_p*float(reward)) ):
					possible_actions[rem_action] = rem_reward
					if(rem_action == 'no_action'):
						print "no action for rem_"+str(i)
			out = self.callPrism(init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, allowed_actions=["no_op"],mean_lat=mean_lat,std_lat=std_lat);
			no_op_action,no_op_reward = self.parsePrismOutput(out,init_num_nodes,vm_types_num_nodes)
			if(no_op_action != 'no_action' and float(no_op_reward) <= float(reward)+(a_p*float(reward)) ):
				possible_actions[no_op_action] = no_op_reward
			if(no_op_action == 'no_action'):
				print "no action for no_op"
			if(len(possible_actions) > 1):
				print "Multiple possible actions: " + str(len(possible_actions))
			min_probability = 1.1
			actions_prob_dict = {}
			for possible_action in possible_actions.keys():
				#print "Rmax PCTL actions: " + str(actions)
				
				out = self.callPrism(init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes, remove_limits_vm_types, cost_spot_step, spot_bid_price, probabilistic=True, allowed_actions=[possible_action],mean_lat=mean_lat,std_lat=std_lat);
				res_output = out.split("Exporting results below:")[1].split("\n");
				for line in res_output:
					matchResult = re.search('[0-9]+\.[0-9]+', line)
					if matchResult:
						result = matchResult.group(0);
						if(float(result) <= min_probability):
							min_probability = float(result)
						if(float(result) in actions_prob_dict.keys()):
							actions_prob_dict[float(result)].append(possible_action)
						else:
							actions_prob_dict[float(result)] = [possible_action]




#			print "Minimum probability: " + str(min_probability)
			if(min_probability == 1.1): #this means actions_prob_dict is empty
#				print "min probability equals 1.1"
				return action+"_"+str(reward)

			self.boundedProbability = True
			if(a_p == 0):
				self.boundedProbability = False
			actions = []
			if(self.boundedProbability):
				for probability in actions_prob_dict.keys():
					if(probability < 0.05):
						actions += actions_prob_dict[probability]
				if(len(actions) <= 1):
					return action+"_"+str(reward)
			else:
				if(len(actions_prob_dict[min_probability]) == 0):
	#				print "None action with minimum probability WTF!"
					return action+"_"+str(reward)

				if(len(actions_prob_dict[min_probability]) == 1):
	#				print "Action with minimum probability: " + str(actions_prob_dict[min_probability][0])
					action = actions_prob_dict[min_probability][0];
					return action+"_"+str(possible_actions[action])
				actions = actions_prob_dict[min_probability]
#				print "More than one actions with minimum probability: " + str(actions)			
			remove_actions = {}
			add_actions = {}
			no_op_actions = {}
			for an_action in actions:
				if('rem' in an_action):
					remove_actions[int(an_action.split('_')[1])] = an_action
				elif('no_op' in an_action):
					no_op_actions[0] = an_action
				elif('add' in an_action):
					add_actions[int(an_action.split('_')[1])] = an_action

			if (prob_policy == 'economy'):
				if(len(remove_actions) > 0):
					action = remove_actions[max(remove_actions.keys())]
				elif(len(no_op_actions) > 0):
					action = 'no_op'
				else:
					action = add_actions[min(add_actions.keys())]
			elif(prob_policy == 'quality'):
				if(len(add_actions) > 0):
					action = add_actions[max(add_actions.keys())]
				elif(len(no_op_actions) > 0):
					action = 'no_op'
				else:
					action = remove_actions[min(remove_actions.keys())]
			elif(prob_policy == 'balance'):
				if(len(no_op_actions) > 0):
					action = 'no_op'
				elif(len(add_actions) > 0):
					action = add_actions[min(add_actions.keys())]
				else:
					action = remove_actions[min(remove_actions.keys())]
			print "Selected action: " + action
			return action+"_"+str(possible_actions[action])
		else:
			return action+"_"+str(reward)


	def parsePrismOutput(self,out,init_num_nodes,vm_types_num_nodes):
#		print out
#		sys.stdout.flush()
		reward = "100000";
		for line in out.split("\n"):
			if "Result: " in line:
				reward = line.split(' ')[1];
				break
			elif "Time for model checking" in line:
				print line
				sys.stdout.flush()

		self.init_state = str(init_num_nodes)+","
		for vm_type in self.utils.vm_types:
			self.init_state += str(vm_types_num_nodes[vm_type][2])+","
#		for vm_type in self.utils.vm_types:
#			if('_sp' in vm_type):
#				self.init_state += "0,"
		self.init_state += "0,2,"
		if(self.utils.spot):
			self.init_state += "-1,"	
		else:
			self.init_state += "0"	
#		for vm_type in self.utils.vm_types:
#			if('_sp' in vm_type):
#				self.init_state += "0,"
		#print "The initail state is: " + str(self.init_state)
		actions = self.findFirstActionsMultiType()
		if(len(actions) > 0):
			try:
			    remove(self.utils.statesFilePath)
			    remove(self.utils.adversariesPath)
			except OSError:
			    pass
#		print "Rmax PCTL actions: " + str(actions)
		if(len(actions) == 1):
			action = actions[0]
			return (action,reward)
		else:
			print "ERROR no action"
			sys.stdout.flush()
			return ("no_action",100000)



	def getTheFasterAction(self, action1, action2):
		if(('add' in action1) and ('add' not in action2)):
			return action1
		elif(('no_op' in action1) and ('rem' in action2)):
			return action1
		elif(('rem' in action1) and 'rem' in action2):
			if( int(action1.split('_')[2]) > int(action2.split('_')[2]) ):
				return action1
			else:
				return action2
		elif(('add' in action1) and ('add' in action2)):
			if( int(action1.split('_')[2]) > int(action2.split('_')[2]) ):
				return action1
			else:
				return action2
		else:
			return action2
				

	def callPrism(self, init_num_nodes, latency, throughput, prob_weights, vm_types_num_nodes={}, remove_limits_vm_types = {}, cost_spot_step={}, spot_bid_price={}, smg=False, probabilistic=False, allowed_actions=[],mean_lat=-1,std_lat=-1):
		latencyConst = "";	
		for i in range(self.minNumVms,self.maxNumVms+1):
			for j in range(0, self.num_of_clusters):
				for k in range(0,len(latency)):
					for l in range(0,len(latency[k])):
						if(latency[k][l][i][j] == {}):
							latency[k][l][i][j] = 0
#						if(i == 8 or i == 10):
#							print "latency_vm_"+str(i)+"_cluster_"+str(j)+"_step_"+str(l)+"="+str(latency[k][l][i][j]);
						latencyConst += "latency_vm_"+str(i)+"_cluster_"+str(j)+"_step_"+str(l)+"="+str(latency[k][l][i][j]);
						if(i <= self.maxNumVms or j < self.num_of_clusters-1 or k < len(latency)-1 or l < len(k)-1):
							latencyConst += ","
		
		throughputConst = "";	
		for i in range(self.minNumVms,self.maxNumVms+1):
			for j in range(0, self.num_of_clusters):
				for k in range(0,len(throughput)):
					for l in range(0,len(latency[k])):
						if(throughput[k][l][i][j] == {}):
							throughput[k][l][i][j] = 0
						throughputConst += "throughput_vm_"+str(i)+"_cluster_"+str(j)+"_step_"+str(l)+"="+str(throughput[k][l][i][j]);
						if(i <= self.maxNumVms or j < self.num_of_clusters-1 or k < len(throughput)-1 or l < len(k)-1):
							throughputConst += ","

		prob_weightsConst = "";	
		for i in range(self.minNumVms,self.maxNumVms+1):
			for j in range(0, self.num_of_clusters):
				for k in range(0,len(prob_weights)):
					for l in range(0,len(latency[k])):
						if(prob_weights[k][l][i][j] == {}):
							prob_weights[k][l][i][j] = 0
#						if(i == 8 or i == 10):
#							print "prob_vm_"+str(i)+"_cluster_"+str(j)+"_step_"+str(l)+"_w="+str(prob_weights[k][l][i][j]);
#							sys.stdout.flush()
						prob_weightsConst += "prob_vm_"+str(i)+"_cluster_"+str(j)+"_step_"+str(l)+"_w="+str(prob_weights[k][l][i][j]);
						if(i <= self.maxNumVms or j < self.num_of_clusters-1 or k < len(prob_weights)-1 or l < len(k)-1):
							prob_weightsConst += ","


		remove_stepsConst = ""
		if(len(remove_limits_vm_types) > 0):
			for vm_type in remove_limits_vm_types:
				cnt = 0
				for remove_step in remove_limits_vm_types[vm_type]:
					remove_stepsConst += "remove"+vm_type+"_step_" + str(cnt) + "=" + str(remove_step);
					if(cnt < len(remove_limits_vm_types[vm_type])):
						remove_stepsConst += ','
					cnt = cnt + 1

		attack_probsConst = ""
		if(self.utils.attack):
			cnt = 1
			for vm_type in self.utils.vm_types_attack_probs.keys():
				attack_probsConst += vm_type+"_attack_prob="+str(self.utils.vm_types_attack_probs[vm_type])
				if(cnt< len(self.utils.vm_types_attack_probs)):
					attack_probsConst += ","
				cnt = cnt + 1
		if(self.utils.multiType):
			vm_type_costsConst = ""
			cnt = 1
			for vm_type in self.utils.vm_types_hour_cost.keys():
				vm_type_costsConst += vm_type+"_cost="+str(self.utils.vm_types_hour_cost[vm_type])
				if(cnt< len(self.utils.vm_types_hour_cost)):
					vm_type_costsConst += ","
				cnt = cnt + 1

			if(len(vm_types_num_nodes)>0):
				vm_types_current_nodesConst = ""
				cnt = 1
				for vm_type in vm_types_num_nodes.keys():
					vm_types_current_nodesConst += vm_type + "_min_num_nodes=" + str(vm_types_num_nodes[vm_type][0]) + ","
					vm_types_current_nodesConst += vm_type + "_max_num_nodes=" + str(vm_types_num_nodes[vm_type][1]) + ","
					vm_types_current_nodesConst += vm_type + "_init_num_nodes=" + str(vm_types_num_nodes[vm_type][2]) + ","
					if(cnt < len(vm_types_num_nodes)):
						vm_types_current_nodesConst += ","
					cnt = cnt + 1
#			remove_spot_stepConst = ""
#			if(len(remove_spot_step)>0):				
#				for vm_type in remove_spot_step.keys():
#					step = 0;
#					for spot_rem in remove_spot_step[vm_type]:
#						remove_spot_stepConst += vm_type + "_remove_spot_step_" + str(step) + "=" + str(spot_rem) + ","
#						step = step + 1	

			cost_spot_stepConst = ""
			if(len(cost_spot_step)>0):				
				for vm_type in cost_spot_step.keys():
					step = 0;
					for spot_rem in cost_spot_step[vm_type]:
						cost_spot_stepConst += vm_type + "_market_price_step_" + str(step) + "=" + str(spot_rem) + ","
						step = step + 1	

			spot_bid_priceConst = ""
			if(len(spot_bid_price)>0):				
				for vm_type in spot_bid_price.keys():
					spot_bid_priceConst += vm_type + "_bid_price" + "=" + str(spot_bid_price[vm_type]) + ","


			mean_latConst = ""
			std_latConst = ""
			if(mean_lat > -1 and std_lat > -1):
				mean_latConst += "mean_lat="+str(mean_lat)
				std_latConst += "std_lat="+str(std_lat)
	

		actionRestrictionsConst = ""
		if("add" in allowed_actions and "remove" in allowed_actions and "no_op" in allowed_actions):
			for i in range(1,self.utils.add_nodes+1):
				actionRestrictionsConst += "add_"+str(i)+"_restriction=false,"
			for i in range(1,self.utils.rem_nodes+1):
				actionRestrictionsConst += "rem_"+str(i)+"_restriction=false,"
			actionRestrictionsConst += "no_op_restriction=false"
		elif("remove" in allowed_actions and "no_op" in allowed_actions):
			for i in range(1,self.utils.add_nodes+1):
				actionRestrictionsConst += "add_"+str(i)+"_restriction=true,"
			for i in range(1,self.utils.rem_nodes+1):
				actionRestrictionsConst += "rem_"+str(i)+"_restriction=false,"
			actionRestrictionsConst += "no_op_restriction=false"
		else:
			for i in range(1,self.utils.add_nodes+1):
				if("add_"+str(i) in allowed_actions):
					actionRestrictionsConst += "add_"+str(i)+"_restriction=false,"
				else:
					actionRestrictionsConst += "add_"+str(i)+"_restriction=true,"
			for i in range(1,self.utils.rem_nodes+1):
				if("rem_"+str(i) in allowed_actions):
					actionRestrictionsConst += "rem_"+str(i)+"_restriction=false,"
				else:
					actionRestrictionsConst += "rem_"+str(i)+"_restriction=true,"
			if("no_op" in allowed_actions):
				actionRestrictionsConst += "no_op_restriction=false"
			else:
				actionRestrictionsConst += "no_op_restriction=true"
			

		if(self.utils.multiType):
#			a =  self.prismFilePath + ' ' + self.modelFilePath + ' ' + ((self.propertiesProbabilisticFilePath if probabilistic else self.propertiesFilePath) if self.pf == None else '-pf '+self.pf) + ' -const init_num_nodes=' + str(init_num_nodes) + ',' + std_latConst + ',' + mean_latConst + ',' + cost_spot_stepConst + spot_bid_priceConst + attack_probsConst + ',' + vm_type_costsConst + ',' + vm_types_current_nodesConst + ',' + latencyConst + ((',' + remove_stepsConst) if len(remove_limits_vm_types) > 0 else "")+ actionRestrictionsConst + ',' + prob_weightsConst + ' -ex -gs -exportresults stdout -exportadv ' + self.utils.adversariesPath + ' -exportstates ' + self.utils.statesFilePath
#			print a

			proc = subprocess.Popen(self.prismFilePath + ' ' + 
							self.modelFilePath + ' ' + 
							((self.propertiesProbabilisticFilePath if probabilistic else self.propertiesFilePath) if self.pf == None else '-pf "'+self.pf.replace('"','\\"')+'"') +
							' -const init_num_nodes=' + str(init_num_nodes) + ',' + 
							std_latConst + ',' + 
							mean_latConst + ',' + 
							cost_spot_stepConst + 
							spot_bid_priceConst +
							attack_probsConst + ',' + 
							vm_type_costsConst + ',' + 
							vm_types_current_nodesConst + ',' + 
							latencyConst + 
							((',' + remove_stepsConst) if len(remove_limits_vm_types) > 0 else "")+ 
							actionRestrictionsConst + ',' + 
							prob_weightsConst + 
							' -ex -gs -exportresults stdout -exportadv ' + 
							self.utils.adversariesPath + 
							' -exportstates ' + 
							self.utils.statesFilePath, stdout=subprocess.PIPE, shell=True);
		else:
			if(probabilistic):
				proc = subprocess.Popen(self.prismFilePath + ' ' + self.modelFilePath + ' ' + (self.propertiesProbabilisticFilePath if self.pf == None else '-pf '+self.pf) + ' -const init_num_nodes=' + str(init_num_nodes) + ',' + latencyConst + ',' + throughputConst + ',' + remove_stepsConst + (',add_restriction=true' if add_restriction else ',add_restriction=false') + ',' + prob_weightsConst + ' -ex -gs -exportresults stdout -exportadv ' + self.utils.adversariesPath + ' -exportstates ' + self.utils.statesFilePath, stdout=subprocess.PIPE, shell=True);
			else:
				#print("PRISM EXECUTION COMMAND: "+self.prismFilePath + ' ' + self.modelFilePath + ' ' + self.propertiesFilePath + ' -const init_num_nodes=' + str(init_num_nodes) + ',serv_throughput=' + str(serv_throughput) + ',' + latencyConst + ((remove_stepsConst) if remove_steps else "") + prob_weightsConst + ' -s -gs -exportresults stdout -exportadv ' + self.utils.adversariesPath + ' -exportstates ' + self.utils.statesFilePath);
				proc = subprocess.Popen(self.prismFilePath + ' ' + self.modelFilePath + ' ' + (self.propertiesFilePath if self.pf == None else '-pf '+self.pf) + ' -const init_num_nodes=' + str(init_num_nodes) + ',' + latencyConst + ((',' + remove_stepsConst) if remove_steps else "")+ (',add_restriction=true' if add_restriction else ',add_restriction=false') + ',' + prob_weightsConst + ' -ex -gs -exportresults stdout -exportadv ' + self.utils.adversariesPath + ' -exportstates ' + self.utils.statesFilePath, stdout=subprocess.PIPE, shell=True);
		(out, err) = proc.communicate()	
		return out

	def findFirstActionsMultiType(self):
		self.startState = self.findInitState();
		#print "the start state is: " + str(self.startState)
		firstActions = [];
		end_state = '';
		for i in xrange(12):
		    try:
			with open(self.utils.adversariesPath) as fin:
				for line in fin:
					matchInit = re.match("("+self.startState+")\s([0-9]+)\s([0-9]+\.?[0-9]*)\s(pre_spot|spot|no_op|add_[0-9]+.*|rem_[0-9]+.*)", line);
					if matchInit:
						start_state = matchInit.group(1);
						end_state = matchInit.group(2);
						action_prob = float(matchInit.group(3));
						action = matchInit.group(4);
						if(not self.utils.spot):
							firstActions.append(action)
					if(end_state != '' and self.utils.spot):
						matchNext = re.match("("+end_state+")\s([0-9]+)\s([0-9]+\.?[0-9]*)\s(pre_spot|spot|no_op|add_[0-9]+.*|rem_[0-9]+.*)", line);
						if matchNext:
							action = matchNext.group(4);
							firstActions.append(action)
						
			break	
		    except IOError:
			time.sleep(3)
		else:
    			raise IOError('Could not access {} after {} attempts'.format(self.utils.adversariesPath, str(12)))		
		return firstActions

	def findFirstActionsFirst(self):
		self.startState = self.findInitState();
		firstActions = [];
		for i in xrange(12):
		    try:
			with open(self.utils.adversariesPath) as fin:
				for line in fin:
					matchInit = re.match("("+self.startState+")\s([0-9]+)\s([0-9]+\.?[0-9]*)\s(no_op|add_[0-9]+|remove_[0-9]+)", line);
					if matchInit:
						start_state = matchInit.group(1);
						end_state = matchInit.group(2);
						action_prob = float(matchInit.group(3));
						action = matchInit.group(4);	
						firstActions.append(action)
						
			break	
		    except IOError:
			time.sleep(3)
		else:
    			raise IOError('Could not access {} after {} attempts'.format(self.utils.adversariesPath, str(12)))		
		return firstActions

	def findFirstActionsSMG(self):
		self.startState = self.findInitState();
		firstActions = [];
		for i in xrange(12):
		    try:
			with open(self.utils.adversariesPath) as fin:
				for line in fin:
					regex = "("+self.startState+")\s([0-9]+)\s(//no_op|//add_[0-9]+_[0-9]+|//remove_[0-9]+_[0-9]+)"
					matchInit = re.match(regex, line);
					if matchInit:
						start_state = matchInit.group(1);
						action = matchInit.group(3);	
						firstActions.append(action.replace('//',''))
						
			break	
		    except IOError:
			time.sleep(3)
		else:
    			raise IOError('Could not access {} after {} attempts'.format(self.utils.adversariesPath, str(12)))
						
		return firstActions

	def findPossibleActions(self):
		self.startState = self.findInitState();
		possibleActions = []
		for i in xrange(12):
		    try:
			with open(self.utils.transitionFilePath) as fin:
				for line in fin:
					regex = "("+self.startState+")\s([0-9]+)\s([0-9]+)\s(.)"
					matchInit = re.match(regex, line);
					if matchInit:
						possibleActions.append(matchInit.group(2));
						
			break	
		    except IOError:
			time.sleep(3)
		else:
    			raise IOError('Could not access {} after {} attempts'.format(self.utils.transitionFilePath, str(12)))
		return len(set(possibleActions))
		

	def findInitState(self):
		for i in xrange(30):
		    try:
			with open(self.utils.statesFilePath) as fin:
				for line in fin:
					if(self.utils.multiType):
						if self.init_state in line:
							return line.split(':')[0]
					else:
						if line.split(',')[1] == '0':
							return line.split(':')[0]
		    	break	
		    except IOError:
			time.sleep(5)
		else:
    			raise IOError('Could not access {} after {} attempts'.format(self.utils.statesFilePath, str(5)))









